#  Copyright 2023 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from synnax.channel.payload import ChannelParams
from synnax.channel.retrieve import ChannelRetriever
from synnax.control.controller import Controller
from synnax.framer import Client as FrameClient
from synnax.telem.control import Authority, CrudeAuthority


class Client:
    framer: FrameClient
    retriever: ChannelRetriever

    def __init__(self, framer: FrameClient, channels: ChannelRetriever) -> None:
        self.framer = framer
        self.retriever = channels

    def acquire(
        self,
        name: str,
        read: ChannelParams,
        write: ChannelParams,
        write_authorities: CrudeAuthority | list[CrudeAuthority] = Authority.ABSOLUTE,
    ) -> Controller:
        return Controller(
            name=name,
            write=write,
            read=read,
            frame_client=self.framer,
            retriever=self.retriever,
            write_authorities=write_authorities,
        )



#  Copyright 2023 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from enum import Enum

from alamos import NOOP, Instrumentation, trace
from freighter import EOF, ExceptionPayload, Payload, Stream, StreamClient

from synnax.channel.payload import ChannelKeys
from synnax.exceptions import UnexpectedError
from synnax.framer.adapter import ReadFrameAdapter
from synnax.framer.frame import Frame, FramePayload
from synnax.telem import TimeRange, TimeSpan, TimeStamp

AUTO_SPAN = TimeSpan(-1)


class _Command(int, Enum):
    OPEN = 0
    NEXT = 1
    PREV = 2
    SEEK_FIRST = 3
    SEEK_LAST = 4
    SEEK_LE = 5
    SEEK_GE = 6
    VALID = 7
    ERROR = 8


class _ResponseVariant(int, Enum):
    NONE = 0
    ACK = 1
    DATA = 2


class _Request(Payload):
    command: _Command
    span: TimeSpan | None = None
    bounds: TimeRange | None = None
    stamp: TimeStamp | None = None
    keys: ChannelKeys | None = None


class _Response(Payload):
    variant: _ResponseVariant
    command: _Command
    ack: bool
    error: str | None
    frame: FramePayload


class Iterator:
    """Used to iterate over a databases telemetry in time-order. It should not be
    instantiated directly, and should instead be instantiated using the segment Client.

    Using an iterator is ideal when querying/processing large ranges of data, but is
    relatively complex and difficult to use. If you're looking to retrieve telemetry
    between two timestamps, see the segment Client read method instead.
    """

    __ENDPOINT = "/frame/iterate"
    __stream: Stream[_Request, _Response]
    __adapter: ReadFrameAdapter

    open: bool
    tr: TimeRange
    instrumentation: Instrumentation
    value: Frame

    def __init__(
        self,
        tr: TimeRange,
        client: StreamClient,
        adapter: ReadFrameAdapter,
        instrumentation: Instrumentation = NOOP,
    ) -> None:
        self.tr = tr
        self.instrumentation = instrumentation
        self.__adapter = adapter
        self.__stream = client.stream(self.__ENDPOINT, _Request, _Response)
        self.__open()

    @trace("debug", "open")
    def __open(self):
        """Opens the iterator, configuring it to iterate over the telemetry in the
        channels with the given keys within the provided time range.

        :param keys: The keys of the channels to iterate over.
        :param tr: The time range to iterate over.
        """
        self._exec(command=_Command.OPEN, bounds=self.tr, keys=self.__adapter.keys)
        self.value = Frame()

    @trace("debug")
    def next(self, span: TimeSpan) -> bool:
        """Reads the next time span of telemetry for each channel in the iterator.

        :param span: The span of time to read. A negative span is equivalent to calling
        prev with the absolute value of the span. If AUTO_SPAN is provided, the iterator
        will automatically determine the span of time to read. This is useful for iterating
        quickly over an entire time range.

        :returns: False if a segment satisfying the request can't be found for a
        particular channel or the iterator has accumulated an error.
        """
        return self._exec(command=_Command.NEXT, span=span)

    @trace("debug")
    def prev(self, span: TimeSpan) -> bool:
        """Reads the previous time span of telemetry for each channel in the iterator.

        :param span: The span of time to read. A negative span is equivalent to calling
        next with the absolute value of the span. If AUTO_SPAN is provided, the iterator
        will automatically determine the span of time to read. This is useful for iterating
        quickly over an entire time range.

        :returns: False if a segment satisfying the request can't be found for a particular
        channel or the iterator has accumulated an error.
        """
        return self._exec(command=_Command.NEXT, span=span)

    @trace("debug")
    def seek_first(self) -> bool:
        """Seeks the iterator to the first segment in the time range, but does not read
        it. Also invalidates the iterator. The iterator will not be considered valid
        until a call to first, last, next, prev, prev_span, next_span, or next_range.

        :returns: False if the iterator is not pointing to a valid segment for a particular
        channel or has accumulated an error.
        """
        return self._exec(command=_Command.SEEK_FIRST)

    @trace("debug")
    def seek_last(self) -> bool:
        """Seeks the iterator to the last segment in the time range, but does not read it.
        Also invalidates the iterator. The iterator will not be considered valid
        until a call to first, last, next, prev, prev_span, next_span, or next_range.

        :returns: False if the iterator is not pointing to a valid segment for a particular
        channel or has accumulated an error.
        """
        return self._exec(command=_Command.SEEK_LAST)

    @trace("debug")
    def seek_lt(self, stamp: TimeStamp) -> bool:
        """Seeks the iterator to the first segment whose start is less than or equal to
        the provided timestamp. Also invalidates the iterator. The iterator will not be
        considered valid until a call to first, last, next, prev, prev_span, next_span, or next_range.

        :returns: False if the iterator is not pointing to a valid segment for a particular
        channel or has accumulated an error.
        """
        return self._exec(command=_Command.SEEK_LE, stamp=stamp)

    @trace("debug")
    def seek_ge(self, stamp: TimeStamp) -> bool:
        """Seeks the iterator to the first segment whose start is greater than or equal to
        the provided timestamp. Also invalidates the iterator. The iterator will not be
        considered valid until a call to first, last, next, prev, prev_span, next_span, or next_range.

        :returns: False if the iterator is not pointing to a valid segment for a particular
        channel or has accumulated an error.
        """
        return self._exec(command=_Command.SEEK_GE, stamp=stamp)

    @trace("debug")
    def valid(self) -> bool:
        """Returns true if the iterator value contains a valid segment, and False otherwise.
        valid most commonly returns false when the iterator is exhausted or has accumulated
        an error.
        """
        return self._exec(command=_Command.VALID)

    @trace("debug")
    def close(self):
        """Close closes the iterator. An iterator MUST be closed after use, and this method
        should probably be placed in a 'finally' block. If the iterator is not closed, it make
        leak resources and threads.
        """
        exc = self.__stream.close_send()
        if exc is not None:
            raise exc
        r, exc = self.__stream.receive()
        if exc is None:
            raise UnexpectedError(
                """Unexpected missing close acknowledgement from server.
                Please report this issue to the Synnax team."""
            )
        elif not isinstance(exc, EOF):
            raise exc

    def __iter__(self):
        self.seek_first()
        return self

    def __next__(self):
        if not self.next(AUTO_SPAN):
            raise StopIteration
        return self.value

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def _exec(self, **kwargs) -> bool:
        exc = self.__stream.send(_Request(**kwargs))
        if exc is not None:
            raise exc
        self.value = Frame()
        while True:
            r, exc = self.__stream.receive()
            if exc is not None:
                raise exc
            assert r is not None
            if r.variant == _ResponseVariant.ACK:
                return r.ack
            fr = Frame(columns_or_data=r.frame.keys, series=r.frame.series)
            self.value.append(self.__adapter.adapt(fr))



#  Copyright 2023 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from __future__ import annotations

from dataclasses import dataclass
from typing import Literal, cast

from freighter import Payload

from synnax.telem import DataType, Rate
from synnax.util.normalize import normalize

ChannelKey = int
ChannelName = str
ChannelKeys = list[int]
ChannelNames = list[str]
ChannelParams = ChannelKeys | ChannelNames | ChannelKey | ChannelName


class ChannelPayload(Payload):
    """A payload container that represent the properties of a channel exchanged to and
    from the Synnax server.
    """

    key: ChannelKey = 0
    data_type: DataType
    rate: Rate = Rate(0)
    name: str = ""
    leaseholder: int = 0
    is_index: bool = False
    index: ChannelKey = 0

    def __str__(self):
        return f"Channel({self.key}, {self.name})"

    def __hash__(self) -> int:
        return hash(self.key)


@dataclass
class NormalizedChannelKeyResult:
    single: bool
    variant: Literal["keys"]
    params: ChannelKeys


@dataclass
class NormalizedChannelNameResult:
    single: bool
    variant: Literal["names"]
    params: ChannelNames


def normalize_channel_params(
    params: ChannelParams,
) -> NormalizedChannelKeyResult | NormalizedChannelNameResult:
    """Determine if a list of keys or names is a single key or name."""
    normalized = normalize(params)
    if len(normalized) == 0:
        return NormalizedChannelKeyResult(single=False, variant="keys", params=[])
    single = isinstance(params, (ChannelKey, ChannelName))
    if isinstance(normalized[0], str):
        return NormalizedChannelNameResult(
            single=single,
            variant="names",
            params=cast(ChannelNames, normalized),
        )
    if isinstance(normalized[0], ChannelPayload):
        return NormalizedChannelNameResult(
            single=single,
            variant="keys",
            params=cast(ChannelNames, [c.key for c in normalized]),
        )
    return NormalizedChannelKeyResult(
        single=single,
        variant="keys",
        params=cast(ChannelKeys, normalized),
    )



#  Copyright 2023 Synnax Labs, Inc.
#
#  Use of this software is governed by the Business Source License included in the file
#  licenses/BSL.txt.
#
#  As of the Change Date specified in that file, in accordance with the Business Source
#  License, use of this software will be governed by the Apache License, Version 2.0,
#  included in the file licenses/APL.txt.

from __future__ import annotations

from typing import Literal, cast, overload

from freighter import Payload
from pandas import DataFrame
from pydantic import Field

from synnax.channel.payload import (
    ChannelKey,
    ChannelKeys,
    ChannelName,
    ChannelNames,
    ChannelParams,
)
from synnax.exceptions import ValidationError
from synnax.telem import Series, TimeRange, TypedCrudeSeries
from synnax.util.normalize import normalize


class FramePayload(Payload):
    keys: ChannelKeys
    series: list[Series]

    def __init__(
        self,
        keys: list[int] | None = None,
        series: list[Series] | None = None,
    ):
        # This is a workaround to allow for a None value to be
        # passed to the arrays field, but still have required
        # type hinting.
        if series is None:
            series = list()
        if keys is None:
            keys = list()
        super().__init__(series=series, keys=keys)


CrudeColumnType = Literal["keys", "names", None]


class ColumnType:
    v: CrudeColumnType

    def __init__(self, v: CrudeColumnType):
        self.v = v

    @classmethod
    def from_channel_params(cls, columns: ChannelParams) -> ColumnType:
        normal = normalize(columns)
        if len(normal) == 0:
            return cls(None)
        first = normal[0]
        if isinstance(first, ChannelKey):
            return cls("keys")
        return cls("names")

    def __eq__(self, rhs: object):
        if rhs is None or rhs == "keys" or rhs == "names":
            c = ColumnType(cast(CrudeColumnType, rhs))
            return self.v is None or c.v is None or c.v == self.v
        return False


class Frame:
    columns: ChannelKeys | ChannelNames
    series: list[Series] = Field(default_factory=list)

    def __new__(cls, *args, **kwargs):
        return super().__new__(cls)
        # return super().__new__(overload_np_array_operators(cls, "to_df"))

    def __init__(
        self,
        columns_or_data: ChannelKeys
        | ChannelNames
        | DataFrame
        | Frame
        | FramePayload
        | dict[ChannelKey, TypedCrudeSeries]
        | None = None,
        series: list[TypedCrudeSeries] | None = None,
    ):
        if isinstance(columns_or_data, Frame):
            self.columns = columns_or_data.columns
            self.series = columns_or_data.series
        elif isinstance(columns_or_data, FramePayload):
            self.columns = columns_or_data.keys
            self.series = columns_or_data.series
        elif isinstance(columns_or_data, DataFrame):
            self.columns = columns_or_data.columns.to_list()
            self.series = [Series(data=columns_or_data[k]) for k in self.columns]
        elif isinstance(columns_or_data, dict):
            self.columns = list(columns_or_data.keys())
            self.series = [Series(d) for d in columns_or_data.values()]
        elif (series is None or isinstance(series, list)) and (
            columns_or_data is None or isinstance(columns_or_data, list)
        ):
            self.series = series or list[Series]()
            self.columns = columns_or_data or list[ChannelKey]()
        else:
            raise ValueError(
                f"""
                [Frame] - invalid construction arguments. Received {columns_or_data}
                and {series}.
            """
            )

    def __str__(self) -> str:
        return self.to_df().__str__()

    @property
    def col_type(self) -> ColumnType:
        return ColumnType.from_channel_params(self.columns)

    @overload
    def append(self, col_or_frame: ChannelKey | ChannelName, array: Series) -> None:
        ...

    @overload
    def append(self, col_or_frame: Frame) -> None:
        ...

    def append(
        self,
        col_or_frame: ChannelKey | ChannelName | Frame,
        array: Series | None = None,
    ) -> None:
        if isinstance(col_or_frame, Frame):
            if self.col_type != col_or_frame.col_type:
                raise ValidationError(
                    f"""
                    Cannot append frame with different label type
                    {self.col_type} != {col_or_frame.col_type}
                """
                )
            self.series.extend(col_or_frame.series)  # type: ignore
            self.columns.extend(col_or_frame.columns)  # type: ignore
        else:
            if array is None:
                raise ValidationError("Cannot append key without array")
            if self.col_type != ColumnType.from_channel_params(
                [col_or_frame]
            ):  # type: ignore
                raise ValidationError("Cannot append array with different label type")
            self.series.append(array)
            self.columns.append(col_or_frame)  # type: ignore

    def items(
        self,
    ) -> list[tuple[ChannelKey, Series]] | list[tuple[ChannelName, Series]]:
        return zip(self.columns, self.series)  # type: ignore

    def __getitem__(self, key: ChannelKey | ChannelName | any) -> Series:
        if not isinstance(key, (ChannelKey, ChannelName)):
            return self.to_df()[key]
        return self.series[self.columns.index(key)]  # type: ignore

    def get(
        self, key: ChannelKey | ChannelName, default: Series | None = None
    ) -> Series | None:
        try:
            return self[key]
        except ValueError:
            return default

    def to_payload(self):
        if self.col_type == "names":
            raise ValidationError(
                "Cannot convert a frame labeled by names to a payload"
            )
        return FramePayload(keys=self.columns, series=self.series)

    def to_df(self) -> DataFrame:
        return DataFrame({k: s for k, s in self.items()})
